
/*
 * splinter 0.1.8
 * 2016 - 2024
 * published under BSD license, see license.txt
 * by Graeme Sheppard http://www.rillion.net/splinter/
 * a reworking and extension of jumpgate by
 * Patroklos Argyroudis http://jumpgate.sourceforge.net/
 */

/*
 * This code does not have header file, because I want it to be small.
 * It isn't very many lines of code. Someone can extend it (IPv6) and
 * put in many other features, call it something else and that is ok.
 * Tell me about it and I may link to it.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <syslog.h>
#include <errno.h>
#include <time.h>


#define nMaxIPchars       20    /* length > 111.222.111.222 */
#define nMaxFD            23    /* above the number id of network FD - unsure how this was arrived at */
#define nTrafficSize    4096    /* length of network data exchange */
#define nBacklogSize    1024    /* size of pending queue */
#define nMaxLine         256    /* length of line of text */
#define nMaxField         32    /* length of field */


typedef enum {eCount, eLabels, eFinal} eCfgPass;
typedef enum {eEmpty = 0, eSingle, eCidr, eRange} eMaskType;
typedef enum {eNormal, eBreak, eContinue} eSubResult;
const char szMasks[] = "0scr"; /* in order as eMaskType above */


struct splinter_record
{
    eMaskType iMaskType;
    unsigned int iFrom;
    unsigned int iTo;
    in_addr_t iDestAddr;
    int iDestPort;
};


struct label_record
{
    char* szLabelName;
    in_addr_t iDestAddr;
    int iDestPort;
};


struct config_data
{
    const char* szFileName;
    struct splinter_record* pSplinters;
    struct label_record* pLabelData;
    char* pLabelNames;
    int iSplinters;
    int iLabelCount;
    int iLabelChars;
    in_addr_t iListenAddr;
    int iListenPort;
    in_addr_t iDefaultAddr;
    int iDefaultPort;
};


struct jumpgate_data
{
    struct in_addr sSourceAddr;
    struct in_addr iServer;
    char szSourceAddr [nMaxIPchars];
    int iFD, iServerFD;
    in_addr_t iServerAddr;
    char szServerAddr [nMaxIPchars];
    int iClientPort, iServerPort;
    struct timeval sStartTime;
    int iTCPerrors;
    char szLogLine [nMaxLine];
    int iReportTimeDiff;
};


struct cfg_data
{
    FILE *fCfg;
    char szLine [nMaxLine];
    char szError [nMaxLine];
    char szName [nMaxField];
    char szValue [nMaxField];
    char szNameLeft [nMaxField];
    char szNameRight [nMaxField];
    char szValueLeft [nMaxField];
    char szValueRight [nMaxField];
    int iNameLeft, iNameRight;
    in_addr_t iDest;
    int iPort;
    char cName, cValue;
    char *szIn;
    char *pcLetter;
    int iLine;
    int iErrors;
    int iReturnCode;
    int iSplinterIndex;
    int iNameLength;
    int iLabelIndex;
    int iLabelStylus;
    int iCIDR, iAdd, iCount;
    int iSysLogCount;
    struct splinter_record* pSplinters;
    struct splinter_record aSplinter;
};


struct cline_options
{
    struct in_addr iTestIP;
    int iDumpCfg;
    int iDaemonize;
    int iTCPerrors;
    int iConfDone;
};


void wipe_string (char *szText)
{
    szText [0] = '\000';
}


void log_message (const char* szFlags, const char* szMessage)
{
    const char szLogLevel[] = "loglevel";
    const char szSysLog[] = "syslog";
    static int iLogLevel = 0;
    static int iSysLog = 0;

    char szOut [nMaxLine];
    int iErrorNum, iLevel;
    const char* pLetter;

    if (strncmp (szMessage, szLogLevel, sizeof (szLogLevel)) == 0)
    {
        iLogLevel = atoi (szFlags);
        return;
    }

    if (strncmp (szMessage, szSysLog, sizeof (szSysLog)) == 0)
    {
        iSysLog = 1;
        return;
    }

    wipe_string (szOut);
    iErrorNum = errno;
    if (!strchr (szFlags, 'f'))
        iErrorNum = 0;

    if (strchr (szFlags, 't'))
    {
        time_t tTime;
        struct tm *pTime;

        tTime = time (NULL);
        pTime = localtime (&tTime);
        strftime (szOut, sizeof (szOut), "%Y-%m-%d %H:%M:%S ", pTime);
    }

    strncat (szOut, szMessage, sizeof (szOut) - 1);

    if (iErrorNum > 0)
    {
        strncat (szOut, ": ", sizeof (szOut) - 1);
        strncat (szOut, strerror (iErrorNum), sizeof (szOut) - 1);
    }

    iLevel = 0;
    pLetter = &szFlags [0];
    while (*pLetter != '\000')
    {
        if (*pLetter >= '0' && *pLetter <= '9')
        {
            iLevel = *pLetter - '0';
            break;
        }
        pLetter++;
    }

    strncat (szOut, "\n", sizeof (szOut) - 1);

    if (strchr (szFlags, 'e'))
        fprintf (stderr, szOut);
    else
    {
        if (iLevel <= iLogLevel)
            printf (szOut);
    }

    if (iSysLog)
      syslog (LOG_NOTICE, szMessage);

    if (strchr (szFlags, 'x'))
        exit (EXIT_SUCCESS);
    if (strchr (szFlags, 'X'))
        exit (EXIT_FAILURE);
}


void syslogstart (const char* szPrefix)
{
    static char szSysLog [nMaxField];
    static int iStatus = 0;

    if (iStatus == 0 && szPrefix != NULL)
    {
        iStatus = 1;
        strncpy (szSysLog, szPrefix, sizeof (szSysLog) - 1);
        return;
    }

    if (iStatus == 1 && szPrefix == NULL)
    {
        openlog (szSysLog, 0, LOG_USER);
        log_message (NULL, "syslog");
        return;
    }
}


struct timeval micro_ticks()
{
    struct timeval sTime;

    if (gettimeofday (&sTime, NULL) == -1)
    {
        sTime.tv_sec = 0;
        sTime.tv_usec = 0;
    }

    return sTime;
}


void text_time_diff (struct timeval sStart, struct timeval sEnd, char* szText, int iSize)
{
    time_t iSeconds;
    suseconds_t iMicros;
    unsigned long iLongSeconds;

    if (sStart.tv_sec <= 0 || sEnd.tv_sec <= 0)
    {
        snprintf (szText, iSize, "time_err");
        return;
    }

    iSeconds = sEnd.tv_sec - sStart.tv_sec;
    if (sEnd.tv_usec >= sStart.tv_usec)
        iMicros = sEnd.tv_usec - sStart.tv_usec;
    else
    {
        iMicros = (sEnd.tv_usec + 1000000) - sStart.tv_usec;
        iSeconds--;
    }
    /* type casting */
    iLongSeconds = iSeconds;
    snprintf (szText, iSize, "%lu.%03d", iLongSeconds, (int) iMicros / 1000);
}


int splinter_filter (in_addr_t sSourceAddr, const struct config_data* pConfig, in_addr_t *dest_addr, int *dest_port)
{
    const struct splinter_record* pSplinters = pConfig->pSplinters;
    struct splinter_record pF;
    eMaskType iType;
    int iFound = 0;
    unsigned int iIP;

    iIP = ntohl ((unsigned int) sSourceAddr);

    for (;;)
    {
        pF = pSplinters [iFound];
        iType = pF.iMaskType;
        if (iType == eEmpty)
        {
            *dest_addr = pConfig->iDefaultAddr;
            *dest_port = pConfig->iDefaultPort;
            return -1;
        }
        if (iType == eSingle && iIP == pF.iFrom)
            break;
        if (iType == eCidr)
        {
            if (pF.iTo == 0) /* catch all */
                break;
            if ((iIP & pF.iTo) == pF.iFrom)
                break;
        }
        if (iType == eRange && iIP >= pF.iFrom && iIP <= pF.iTo)
            break;
        iFound++;
    }

    pF = pSplinters [iFound];
    *dest_addr = pF.iDestAddr;
    *dest_port = pF.iDestPort;
    return iFound;
}


void trim_spaces (char* szText)
{
    char szReduce [nMaxLine];
    int iLength, iIndex;
    char cLetter;

    iLength = strlen (szText);
    if (iLength == 0)
        return;

    iIndex = 0;
    for (;;)
    {
        cLetter = szText [iIndex];
        if (cLetter != ' ')
            break;
        iIndex++;
    }

    if (iIndex > 0)
    {
        snprintf (szReduce, nMaxLine, "%s", szText + iIndex);
        snprintf (szText, nMaxLine, "%s", szReduce);
    }

    iLength = strlen (szText);
    if (iLength == 0)
        return;

    iIndex = iLength - 1;
    for (;;)
    {
        if (iIndex < 0)
            break;
        cLetter = szText [iIndex];
        if (cLetter != ' ')
            break;
        iIndex--;
    }

    szText [iIndex + 1] = '\000';
}


unsigned int atoip (const char *szText)
{
    in_addr_t iAddr;
    int iReturnCode;

    iReturnCode = inet_pton (AF_INET, szText, &iAddr);
    if (iReturnCode != 1)
        return 0;
    else
        return ntohl ((unsigned int) iAddr);
}


void socket_to_text (struct in_addr iAddr, int iPort, char* szText, int iSize)
{
    if (iPort == 0)
        snprintf (szText, iSize, "0");
    else
        snprintf (szText, iSize, "%s:%d", inet_ntoa (iAddr), iPort);
}


void split_delimiter (char cDelimiter, char *pcFound, char *szText, char *szLeft, char *szRight)
{
    char *pcLetter;

    pcLetter = index (szText, cDelimiter);
    if (pcLetter == NULL)
        return;

    *pcFound = cDelimiter;
    *pcLetter = '\000';
    snprintf (szLeft, nMaxLine, szText);
    snprintf (szRight, nMaxLine, pcLetter + 1);
    wipe_string (szText);
}


void fill_splinter (struct splinter_record* pSplinter, eMaskType iMask, int iFrom, int iTo, int iDestAddr, int iDestPort)
{
    pSplinter->iMaskType = iMask;
    pSplinter->iFrom = iFrom;
    pSplinter->iTo = iTo;
    pSplinter->iDestAddr = iDestAddr;
    pSplinter->iDestPort = iDestPort;
}


void zero_splinter (struct splinter_record* pSplinter)
{
    fill_splinter (pSplinter, eEmpty, 0, 0, 0, 0);
}


eSubResult cfg_condense (struct cfg_data* c)
{
    zero_splinter (&c->aSplinter);
    c->iLine++;
    c->szIn = fgets (c->szLine, sizeof (c->szLine), c->fCfg);
    if (c->szIn == NULL)
        return eBreak;

    c->pcLetter = strchr (c->szLine, '\n');
    if (c->pcLetter)
        *c->pcLetter = '\000';
    c->pcLetter = strchr (c->szLine, '#');
    if (c->pcLetter)
        *c->pcLetter = '\000';

    for (;;)
    {
        c->pcLetter = strchr (c->szLine, '\t');
        if (c->pcLetter == NULL)
            break;
        *c->pcLetter = ' ';
    }

    trim_spaces (c->szLine);
    if (strlen (c->szLine) == 0)
        return eContinue;

    c->pcLetter = index (c->szLine, '=');
    if (c->pcLetter == NULL)
    {
        snprintf (c->szError, nMaxLine, "no = setting");
        return eContinue;
    }

    *c->pcLetter = '\000';
    snprintf (c->szName, nMaxLine, c->szLine);
    snprintf (c->szValue, nMaxLine, c->pcLetter + 1);
    wipe_string (c->szNameLeft);
    wipe_string (c->szNameRight);
    wipe_string (c->szValueLeft);
    wipe_string (c->szValueRight);
    c->cName = c->cValue = '.';

    split_delimiter ('/', &c->cName, c->szName, c->szNameLeft, c->szNameRight);
    split_delimiter ('-', &c->cName, c->szName, c->szNameLeft, c->szNameRight);
    split_delimiter ('!', &c->cName, c->szName, c->szNameLeft, c->szNameRight);
    split_delimiter (':', &c->cValue, c->szValue, c->szValueLeft, c->szValueRight);
    split_delimiter ('!', &c->cValue, c->szValue, c->szValueLeft, c->szValueRight);

    trim_spaces (c->szName);
    trim_spaces (c->szValue);
    trim_spaces (c->szNameLeft);
    trim_spaces (c->szNameRight);
    trim_spaces (c->szValueLeft);
    trim_spaces (c->szValueRight);

    c->iDest = 0;
    c->iPort = -1;

    return eNormal;
}


eSubResult cfg_value (struct cfg_data* c, int iPass, struct config_data* pConfig)
{
    if (c->cValue == ':')
    {
        if (strncmp (c->szName, "listen", nMaxLine) == 0 && strncmp (c->szValueLeft, "", nMaxLine) == 0)
            c->iDest = htonl (INADDR_ANY); /* all network interfaces */
        else
        {
            c->iReturnCode = inet_pton (AF_INET, c->szValueLeft, &c->iDest);
            if (c->iReturnCode != 1)
            {
                snprintf (c->szError, nMaxLine, "bad ip address");
                return eContinue;
            }
        }
        c->iPort = atoi (c->szValueRight);
        if (c->iPort <= 0 || c->iPort >= 65536)
        {
            snprintf (c->szError, nMaxLine, "bad destination port");
            return eContinue;
        }
        wipe_string (c->szValueLeft);
        wipe_string (c->szValueRight);
    }

    if (c->cValue == '!' && (iPass == eLabels || iPass == eFinal))
    {
        int iIndex = 0;
        int iLimit = iPass == eFinal ? pConfig->iLabelCount : c->iLabelIndex;
        for (;;)
        {
            if (iIndex >= iLimit)
            {
                iIndex = -1;
                break;
            }
            if (strncmp (pConfig->pLabelData [iIndex].szLabelName, c->szValueRight, sizeof (c->szValueRight)) == 0)
                break;
            iIndex++;
        }
        if (iIndex < 0)
        {
            if (iPass == eFinal)
            {
                snprintf (c->szError, nMaxLine, "label not found");
                return eContinue;
            }
            c->iPort = 0; /* grace for eLabels pass */
        }
        else
        {
            c->iDest = pConfig->pLabelData [iIndex].iDestAddr;
            c->iPort = pConfig->pLabelData [iIndex].iDestPort;
        }
    }

    return eNormal;
}


eSubResult cfg_syslog (struct cfg_data* c, int iPass)
{
    if (strncmp (c->szName, "syslog", nMaxLine) == 0)
    {
        if (c->iSysLogCount > 0)
        {
            snprintf (c->szError, nMaxLine, "more than one syslog");
            return eContinue;
        }
        c->iSysLogCount++;
        if (strncmp (c->szValue, "", nMaxLine) == 0)
        {
            snprintf (c->szError, nMaxLine, "no syslog value");
            return eContinue;
        }
        if (iPass == eFinal)
        {
            syslogstart (c->szValue);
        }
        return eContinue;
    }
    return eNormal;
}


eSubResult cfg_specials (struct cfg_data* c, int iPass, struct config_data* pConfig)
{
    if (strncmp (c->szValue, "0", nMaxLine) == 0)
        c->iPort = 0;

    if ((iPass == eLabels || iPass == eFinal) && c->iPort < 0)
    {
        snprintf (c->szError, nMaxLine, "bad destination");
        return eContinue;
    }

    if (strncmp (c->szName, "listen", nMaxLine) == 0)
    {
        if (c->cValue == '!')
        {
            snprintf (c->szError, nMaxLine, "cannot listen on a label");
            return eContinue;
        }
        pConfig->iListenAddr = c->iDest;
        pConfig->iListenPort = c->iPort;
        return eContinue;
    }

    if (strncmp (c->szName, "default", nMaxLine) == 0)
    {
        pConfig->iDefaultAddr = c->iDest;
        pConfig->iDefaultPort = c->iPort;
        return eContinue;
    }

    return eNormal;
}


eSubResult cfg_label (struct cfg_data* c, int iPass, struct config_data* pConfig)
{
    if (c->cName == '!')
    {
        c->iNameLength = strlen (c->szNameRight);
        if (c->iNameLength <= 0)
        {
            snprintf (c->szError, nMaxLine, "label has no characters");
            return eContinue;
        }
        if (iPass == eCount)
        {
            pConfig->iLabelCount++;
            pConfig->iLabelChars += c->iNameLength + 1; /* add one for trailing 0x00 */
        }
        else if (iPass == eLabels)
        {
            if (c->iLabelIndex > pConfig->iLabelCount)
            {
                snprintf (c->szError, nMaxLine, "label count exceeded");
                return eContinue;
            }
            if (c->iLabelStylus + c->iNameLength + 1 > pConfig->iLabelChars)
            {
                snprintf (c->szError, nMaxLine, "label name overflow");
                return eContinue;
            }
            int iIndex;
            for (iIndex = 0; iIndex < c->iLabelIndex; iIndex++)
            {
                if (strncmp (c->szNameRight, pConfig->pLabelData [iIndex].szLabelName, sizeof (c->szNameRight)) == 0)
                {
                    snprintf (c->szError, nMaxLine, "label \"%s\" repeated", c->szNameRight);
                    return eContinue;
                }
            }
            memcpy (pConfig->pLabelNames + c->iLabelStylus, c->szNameRight, c->iNameLength + 1); /* memcpy */
            pConfig->pLabelData [c->iLabelIndex].szLabelName = pConfig->pLabelNames + c->iLabelStylus;
            pConfig->pLabelData [c->iLabelIndex].iDestAddr = c->iDest;
            pConfig->pLabelData [c->iLabelIndex].iDestPort = c->iPort;
            c->iLabelIndex++;
            c->iLabelStylus += c->iNameLength + 1;
        }
        return eContinue;
    }

    return eNormal;
}


eSubResult cfg_ranges (struct cfg_data* c)
{
    if (c->cName == '.') /* just an IP address then */
    {
        c->iNameLeft = atoip (c->szName);
        if (c->iNameLeft != 0)
        {
            fill_splinter (&c->aSplinter, eSingle, c->iNameLeft, 0, c->iDest, c->iPort);
            return eContinue;
        }
        else
        {
            snprintf (c->szError, nMaxLine, "bad IP address");
            return eContinue;
        }
    }
    else if (c->cName == '/')
    {
        if (strncmp (c->szNameLeft, "0", nMaxLine) == 0 && strncmp (c->szNameRight, "0", nMaxLine) == 0)
        /* the catch all */
        {
            fill_splinter (&c->aSplinter, eCidr, 0, 0, c->iDest, c->iPort);
            return eContinue;
        }
        c->iNameLeft = atoip (c->szNameLeft);
        c->iNameRight = atoip (c->szNameRight);
        if (c->iNameRight == 0)
        {
            c->iCIDR = atoi (c->szNameRight);
            if (c->iCIDR >= 1 && c->iCIDR <= 32)
            {
                c->iNameRight = 0x80000000;
                c->iAdd = 0x40000000;
                for (c->iCount = 2; c->iCount <= c->iCIDR; c->iCount++)
                {
                    c->iNameRight += c->iAdd;
                    c->iAdd = c->iAdd >> 1;
                }
            }
        }
        if (c->iNameLeft == 0 || c->iNameRight == 0)
        {
            snprintf (c->szError, nMaxLine, "bad cidr expression");
            return eContinue;
        }
        fill_splinter (&c->aSplinter, eCidr, c->iNameLeft, c->iNameRight, c->iDest, c->iPort);
        return eContinue;
    }
    else if (c->cName == '-')
    {
        c->iNameLeft = atoip (c->szNameLeft);
        c->iNameRight = atoip (c->szNameRight);
        if (c->iNameLeft == 0 || c->iNameRight == 0 || c->iNameLeft >= c->iNameRight)
        {
            snprintf (c->szError, nMaxLine, "bad ip range");
            return eContinue;
        }
        fill_splinter (&c->aSplinter, eRange, c->iNameLeft, c->iNameRight, c->iDest, c->iPort);
        return eContinue;
    }

    return eNormal;
}


void cfg_file (eCfgPass iPass, struct config_data* pConfig)
{
    eSubResult iAction;
    struct cfg_data cdData;
    struct cfg_data* c = &cdData;

    c->iErrors = 0;
    c->iSplinterIndex = 0;
    c->iLabelIndex = 0;
    c->iLabelStylus = 0;
    c->iSysLogCount = 0;
    c->pSplinters = pConfig->pSplinters;

    c->fCfg = fopen (pConfig->szFileName, "r");
    if (c->fCfg == NULL)
    {
        snprintf (c->szLine, sizeof (c->szLine), "cannot open file \"%s\"", pConfig->szFileName);
        log_message ("feX", c->szLine);
    }
    c->aSplinter.iMaskType = eEmpty; /* '0' = empty */
    c->iLine = 0;
    wipe_string (c->szError);

    for (;;)
    {
        if (strncmp (c->szError, "", nMaxLine) != 0)
        {
            printf ("line %d: %s\n", c->iLine, c->szError);
            c->iErrors = 1;
            wipe_string (c->szError);
        }

        if (c->aSplinter.iMaskType != eEmpty)
        {
            switch (iPass)
            {
                case eCount:
                    c->iSplinterIndex++;
                    break;

                case eLabels:
                    break;

                case eFinal:
                    fill_splinter (&c->pSplinters [c->iSplinterIndex], c->aSplinter.iMaskType,
                        c->aSplinter.iFrom, c->aSplinter.iTo, c->aSplinter.iDestAddr, c->aSplinter.iDestPort);
                    c->iSplinterIndex++;
                    break;
            }
        }

        iAction = cfg_condense (c);
        if (iAction == eBreak)
            break;
        if (iAction == eContinue)
            continue;

        iAction = cfg_value (c, iPass, pConfig);
        if (iAction == eBreak)
            break;
        if (iAction == eContinue)
            continue;

        iAction = cfg_syslog (c, iPass);
        if (iAction == eBreak)
            break;
        if (iAction == eContinue)
            continue;

        iAction = cfg_specials (c, iPass, pConfig);
        if (iAction == eBreak)
            break;
        if (iAction == eContinue)
            continue;

        iAction = cfg_label (c, iPass, pConfig);
        if (iAction == eBreak)
            break;
        if (iAction == eContinue)
            continue;

        iAction = cfg_ranges (c);
        if (iAction == eBreak)
            break;
        if (iAction == eContinue)
            continue;

        snprintf (c->szError, nMaxLine, "general syntax error");
    }

    fclose (c->fCfg);
    if (c->iErrors)
    {
        snprintf (c->szLine, sizeof (c->szLine), "errors in file \"%s\"", pConfig->szFileName);
        log_message ("eX", c->szLine);
    }

    switch (iPass)
    {
        case eCount:
            pConfig->iSplinters = c->iSplinterIndex;
            break;

        case eLabels:
            break;

        case eFinal:
            zero_splinter (&c->pSplinters [c->iSplinterIndex]);
            break;
    }
}


void log_line_text (struct jumpgate_data* pJumpData)
{
    struct sockaddr_in sLocal;
    socklen_t iLocalSize;
    char szLocal [nMaxIPchars];

    iLocalSize = sizeof (sLocal);
    if (getsockname (pJumpData->iServerFD, (struct sockaddr *) &sLocal, &iLocalSize) == -1)
    {
        snprintf (pJumpData->szLogLine, sizeof (pJumpData->szLogLine), "Internal getsockname error");
        return;
    }

    strncpy (szLocal, inet_ntoa (sLocal.sin_addr), sizeof (szLocal) - 1);

    snprintf (pJumpData->szLogLine, sizeof (pJumpData->szLogLine), "%s:%d -> %s:%d -> %s:%d",
        pJumpData->szSourceAddr, pJumpData->iClientPort,
        szLocal, ntohs (sLocal.sin_port),
        pJumpData->szServerAddr, pJumpData->iServerPort);
}


void dump_cfg (struct config_data* sConfig, struct in_addr* iTestIP)
{
    struct in_addr iAddr;

    iAddr.s_addr = sConfig->iListenAddr;
    printf ("Listen: %s:%d\n", inet_ntoa (iAddr), sConfig->iListenPort);
    iAddr.s_addr = sConfig->iDefaultAddr;
    printf ("Default: %s:%d\n", inet_ntoa (iAddr), sConfig->iDefaultPort);
    char szDest [nMaxLine];
    int iIndex;
    for (iIndex = 0; iIndex < sConfig->iLabelCount; iIndex++)
    {
        iAddr.s_addr = sConfig->pLabelData [iIndex].iDestAddr;
        socket_to_text (iAddr, sConfig->pLabelData [iIndex].iDestPort, szDest, sizeof (szDest));
        printf ("!%s = %s\n", sConfig->pLabelData [iIndex].szLabelName, szDest);
    }

    const struct splinter_record *pF;
    struct in_addr iAddrStart;
    struct in_addr iAddrEnd;
    struct in_addr iAddrDest;
    char szIncoming [50];
    char szRangeEnd [30];
    for (iIndex = 0; iIndex < sConfig->iSplinters; iIndex++)
    {
        pF = &sConfig->pSplinters [iIndex];
        iAddrStart.s_addr = ntohl (pF->iFrom);
        iAddrEnd.s_addr = ntohl (pF->iTo);
        iAddrDest.s_addr = pF->iDestAddr;
        switch (pF->iMaskType)
        {
            case eSingle:
                snprintf (szIncoming, sizeof (szIncoming), "%s", inet_ntoa (iAddrStart));
                break;
            case eCidr:
                if (pF->iTo == 0)
                    snprintf (szIncoming, sizeof (szIncoming), "0/0");
                else
                    snprintf (szIncoming, sizeof (szIncoming), "%s/%08x", inet_ntoa (iAddrStart), pF->iTo);
                break;
            case eRange:
                snprintf (szRangeEnd, sizeof (szRangeEnd), "%s", inet_ntoa (iAddrEnd)); /* code needed over two... */
                snprintf (szIncoming, sizeof (szIncoming), "%s-%s", inet_ntoa (iAddrStart), szRangeEnd); /* ... lines */
                break;
            default:
                snprintf (szIncoming, sizeof (szIncoming), "?");
                break;
        }
        socket_to_text (iAddrDest, pF->iDestPort, szDest, sizeof (szDest));
        printf ("%d: %s -> %s\n", iIndex, szIncoming, szDest);
    }

    if (iTestIP->s_addr != 0)
    {
        char szMatch [nMaxLine];
        in_addr_t iDest;
        int iPort;
        struct in_addr sDest;
        int iMatch = splinter_filter (iTestIP->s_addr, sConfig, &iDest, &iPort);
        if (iMatch < 0)
            snprintf (szMatch, sizeof (szMatch), "default");
        else
            snprintf (szMatch, sizeof (szMatch), "%d", iMatch);
        sDest.s_addr = iDest;
        socket_to_text (sDest, iPort, szDest, sizeof (szDest));
        printf ("IP test match=%s: %s -> %s\n", szMatch, inet_ntoa (*iTestIP), szDest);
    }
}


void tcp_exit (const struct jumpgate_data* pJumpData, const char* szSection, const char* szError)
{
    int iErrorNum = 0;
    char szMessage [nMaxLine];
    char szPrintTime [30];
    char szTimeBrackets [60];
    char szExtendedError [nMaxLine - 1];

    if (szError)
        iErrorNum = errno;

    strncpy (szMessage, pJumpData->szLogLine, sizeof (szMessage));

    if (pJumpData->iReportTimeDiff)
    {
        text_time_diff (pJumpData->sStartTime, micro_ticks(), szPrintTime, sizeof (szPrintTime));
        snprintf (szTimeBrackets, sizeof (szTimeBrackets), " (%ss)", szPrintTime);
        strncat (szMessage, szTimeBrackets, sizeof (szMessage) - 1);
    }

    if (pJumpData->iTCPerrors)
    {
        if (szError)
        {
            if (iErrorNum > 0)
                snprintf (szExtendedError, sizeof (szExtendedError), " (%s: %s in %s)", szError, strerror (iErrorNum), szSection);
            else
                snprintf (szExtendedError, sizeof (szExtendedError), " (%s in %s)", szError, szSection);
        }
        else
            snprintf (szExtendedError, sizeof (szExtendedError), " (in %s)", szSection);
    }
    else
        wipe_string (szExtendedError);

    strncat (szMessage, szExtendedError, sizeof (szMessage) - 1);
    log_message ("t5x", szMessage);
}


void tcp_jumpgate (struct jumpgate_data sJumpData)
{
    int iReadChars;
    char buf [nTrafficSize];
    fd_set setMonitor1, setMonitor2;
    struct sockaddr_in sServer;

    if ((sJumpData.iServerFD = socket (AF_INET, SOCK_STREAM, 0)) < 0)
        tcp_exit (&sJumpData, "socket", "socket acquire error");

    sServer.sin_family = AF_INET;
    sServer.sin_port = htons (sJumpData.iServerPort);
    memcpy (&sServer.sin_addr, &sJumpData.iServerAddr, sizeof (in_addr_t));
    sJumpData.iReportTimeDiff = 0;

    if ((connect (sJumpData.iServerFD, (struct sockaddr *) &sServer, sizeof (sServer))) < 0)
    {
        snprintf (sJumpData.szLogLine, sizeof (sJumpData.szLogLine),
            "%s:%d -> connection creation error to %s:%d",
            sJumpData.szSourceAddr, sJumpData.iClientPort,
            sJumpData.szServerAddr, sJumpData.iServerPort);
        tcp_exit (&sJumpData, "connect", "connect error");
    }

    log_line_text (&sJumpData);
    log_message ("t5", sJumpData.szLogLine);
    sJumpData.iReportTimeDiff = 1;

    for (;;)
    {
        FD_ZERO (&setMonitor1);
        FD_ZERO (&setMonitor2);
        FD_SET (sJumpData.iFD, &setMonitor1);
        FD_SET (sJumpData.iFD, &setMonitor2);
        FD_SET (sJumpData.iServerFD, &setMonitor1);
        FD_SET (sJumpData.iServerFD, &setMonitor2);

        if (select (nMaxFD, &setMonitor1, NULL, &setMonitor2, NULL) == -1)
        {
            if ((shutdown (sJumpData.iServerFD, 2)) == -1)
                tcp_exit (&sJumpData, "select", "shutdown error");
            if ((close (sJumpData.iServerFD)) == -1)
                tcp_exit (&sJumpData, "select", "close error");
            tcp_exit (&sJumpData, "select", NULL);
        }

        if (FD_ISSET (sJumpData.iFD, &setMonitor1) || FD_ISSET (sJumpData.iFD, &setMonitor2))
        {
            if ((iReadChars = read (sJumpData.iFD, buf, nTrafficSize)) <= 0)
            {
                if ((shutdown (sJumpData.iServerFD, 2)) == -1)
                    tcp_exit (&sJumpData, "client read", "shutdown error");
                if ((close (sJumpData.iServerFD)) == -1)
                    tcp_exit (&sJumpData, "client read", "close error");
                tcp_exit (&sJumpData, "client read", NULL);
            }
            if ((write (sJumpData.iServerFD, buf, iReadChars)) <= 0)
            {
                if ((shutdown (sJumpData.iServerFD, 2)) == -1)
                    tcp_exit (&sJumpData, "server write", "shutdown error");
                if ((close (sJumpData.iServerFD)) == -1)
                    tcp_exit (&sJumpData, "server write", "close error");
                tcp_exit (&sJumpData, "server write", NULL);
            }
        }
        else if (FD_ISSET (sJumpData.iServerFD, &setMonitor1) || FD_ISSET (sJumpData.iServerFD, &setMonitor2))
        {
            if ((iReadChars = read (sJumpData.iServerFD, buf, nTrafficSize)) < 0)
            {
                if ((shutdown (sJumpData.iServerFD, 2)) == -1)
                    tcp_exit (&sJumpData, "server read", "shutdown error");
                if ((close (sJumpData.iServerFD)) == -1)
                    tcp_exit (&sJumpData, "server read", "close error");
                tcp_exit (&sJumpData, "server read", NULL);
            }
            if ((write (sJumpData.iFD, buf, iReadChars)) <= 0)
            {
                if ((shutdown (sJumpData.iServerFD, 2)) == -1)
                    tcp_exit (&sJumpData, "client write", "shutdown error");
                if ((close (sJumpData.iServerFD)) == -1)
                    tcp_exit (&sJumpData, "client write", "close error");
                tcp_exit (&sJumpData, "client write", NULL);
            }
        }
    }
}


void branch_tcp (int iLocalFD, int iConnectionFD, int iTCPerrors, struct sockaddr_in* sClientSocket, struct config_data* sConfig)
{
    struct jumpgate_data sJumpData;
    in_addr_t sServerAddr;
    int iServerPort;
    char szLogMessage [nMaxLine];
    struct in_addr sDestAddr;

    wipe_string (sJumpData.szLogLine);
    sJumpData.iReportTimeDiff = 0;

    if ((close (iLocalFD)) == -1)
        log_message ("feX", "close error"); /* might be logging only though */

    splinter_filter (sClientSocket->sin_addr.s_addr, sConfig, &sServerAddr, &iServerPort);

    sJumpData.sSourceAddr = sClientSocket->sin_addr;
    strncpy (sJumpData.szSourceAddr, inet_ntoa (sClientSocket->sin_addr), sizeof (sJumpData.szSourceAddr) - 1);
    sJumpData.iClientPort = ntohs (sClientSocket->sin_port);

    if (iServerPort <= 0) /* tear down connection */
    {
        snprintf (szLogMessage, nMaxLine, "%s:%d -> 0", sJumpData.szSourceAddr, sJumpData.iClientPort);
        log_message ("t5", szLogMessage);
        if ((shutdown (iConnectionFD, 2)) == -1)
            tcp_exit (&sJumpData, "client teardown", "shutdown error");
        if ((close (iConnectionFD)) == -1)
            tcp_exit (&sJumpData, "client teardown", "close error");
        exit (EXIT_SUCCESS);
    }

    sJumpData.iFD = iConnectionFD;
    sJumpData.iServerAddr = sServerAddr;
    sDestAddr.s_addr = sServerAddr;
    strncpy (sJumpData.szServerAddr, inet_ntoa (sDestAddr), sizeof (sJumpData.szServerAddr) - 1);
    sJumpData.iServerPort = iServerPort;
    sJumpData.sStartTime = micro_ticks();
    sJumpData.iTCPerrors = iTCPerrors;

    tcp_jumpgate (sJumpData);

    exit (EXIT_SUCCESS); /* added to be sure of child exit */
}


void help()
{
    printf ("spilinter 0.1.8\n");
    printf ("http://www.rillion.net/splinter\n");
    printf ("-f <configuration.file> (mandatory)\n");
    printf ("-t <ip> test ip address\n");
    printf ("-l dump listing\n");
    printf ("-d daemonize\n");
    printf ("-e tcp error data\n");
}


void command_line_options (int argc, char* argv[], struct cline_options* sOpts, struct config_data* sConfig)
{
    char iLetter;

    while ((iLetter = getopt (argc, argv, "vldef:t:")) > 0 && iLetter < 128)
    {
        switch (iLetter)
        {
            case 'v':
                help();
                exit (EXIT_SUCCESS);
            case 'l':
                sOpts->iDumpCfg = 1;
                break;
            case 'd':
                sOpts->iDaemonize = 1;
                break;
            case 'e':
                sOpts->iTCPerrors = 1;
                break;
            case 'f':
                if (sOpts->iConfDone != 0)
                    log_message ("eX", "error: extra config file specified");
                sConfig->szFileName = optarg;
                sConfig->pSplinters = NULL;
                sConfig->pLabelData = NULL;
                sConfig->pLabelNames = NULL;
                sConfig->iSplinters = 0;
                sConfig->iLabelCount = 0;
                sConfig->iLabelChars = 0;
                sConfig->iDefaultPort = -1;
                cfg_file (eCount, sConfig);
                sConfig->pSplinters = (struct splinter_record*) calloc (sConfig->iSplinters + 1, sizeof (struct splinter_record));
                if (sConfig->pSplinters == NULL)
                    log_message ("eX", "error alocating memory to splinters");
                if (sConfig->iLabelCount > 0)
                {
                    sConfig->pLabelData = calloc (sConfig->iLabelCount, sizeof (struct label_record));
                    if (sConfig->pLabelData == NULL)
                        log_message ("eX", "error alocating memory to label data");
                    sConfig->pLabelNames = calloc (1, sConfig->iLabelChars);
                    if (sConfig->pLabelNames == NULL)
                        log_message ("eX", "error alocating memory to label names");
                }
                cfg_file (eLabels, sConfig);
                cfg_file (eFinal, sConfig);
                sOpts->iConfDone = 1;
                break;
            case 't':
                if (sOpts->iTestIP.s_addr != 0)
                    log_message ("eX", "error: may only test one IP");
                sOpts->iTestIP.s_addr = ntohl (atoip (optarg));
                if (sOpts->iTestIP.s_addr == 0)
                    log_message ("eX", "error: bad IP to test");
                sOpts->iDumpCfg = 1;
                break;
        }
    }
}


void child_signal (int iSignal)
{
    /*
     * Wait for all dead processes.
     * We use a non-blocking call to be sure this signal handler will not
     * block if a child was cleaned up in another part of the program.
     */

    while (waitpid (-1, NULL, WNOHANG) > 0)
    {
        /* do nothing */;
    }
}


int daemonize()
{
    pid_t iPID;

    if ((signal (SIGTTOU, SIG_IGN)) == SIG_ERR)
        log_message ("feX", "daemonize signal error");

    if ((signal (SIGTTIN, SIG_IGN)) == SIG_ERR)
        log_message ("feX", "daemonize signal error");

    if ((signal (SIGTSTP, SIG_IGN)) == SIG_ERR)
        log_message ("feX", "daemonize signal error");

    if ((signal (SIGHUP, SIG_IGN)) == SIG_ERR)
        log_message ("feX", "daemonize signal error");

    if ((iPID = fork()) < 0)
        return -1;
    else if (iPID != 0)
        exit (EXIT_SUCCESS);

    if (setsid() == -1)
        log_message ("feX", "setsid error");

    umask (0);
    return 0;
}


int main (int argc, char *argv[])
{
    struct cline_options sOpts;
    struct config_data sConfig;
    int iLocalFD;
    const int iInt = 1;
    pid_t iChildPID;
    socklen_t iSocketSize;
    struct sockaddr_in sLocalSocket, sClientSocket;
    int iConnectionFD;
    struct sigaction sSigAction;

    if (argc <= 1)
    {
        help();
        exit (EXIT_SUCCESS);
    }

    sOpts.iTestIP.s_addr = 0;
    sOpts.iConfDone = 0;
    sOpts.iDumpCfg = 0;
    sOpts.iDaemonize = 0;
    sOpts.iTCPerrors = 0;

    command_line_options (argc, argv, &sOpts, &sConfig);

    if (sOpts.iConfDone == 0)
        log_message ("eX", "no configuration file specified");
    if (sConfig.iListenPort == 0)
        log_message ("eX", "no listen port specified");
    if (sConfig.iListenPort < 1024 && getuid() != 0)
        log_message ("eX", "error: you're trying to listen on a low port without having root privileges");
    if (sConfig.iDefaultPort < 0)
        log_message ("eX", "no default destination specified");

    if (sConfig.iLabelCount > 0)
    {
        free (sConfig.pLabelData);
        free (sConfig.pLabelNames);
    }

    if (sOpts.iDumpCfg)
    {
        dump_cfg (&sConfig, &sOpts.iTestIP);
        free (sConfig.pSplinters);
        exit (EXIT_SUCCESS);
    }

    syslogstart (NULL);

    if (sOpts.iDaemonize)
    {
        if (daemonize() == -1)
            log_message ("eX", "fork error");
        log_message ("0", "loglevel");
    }
    else
        log_message ("5", "loglevel");

    if ((iLocalFD = socket (AF_INET, SOCK_STREAM, 0)) < 0)
        log_message ("feX", "socket error");

    bzero (&sLocalSocket, sizeof (sLocalSocket));
    sLocalSocket.sin_family = AF_INET;
    sLocalSocket.sin_addr.s_addr = sConfig.iListenAddr;
    sLocalSocket.sin_port = htons (sConfig.iListenPort);

    /* if you use linux beware of the SO_REUSEADDR vulnerability,
     * although I think new linux versions fix this problem
     */
    if (setsockopt (iLocalFD, SOL_SOCKET, SO_REUSEADDR, &iInt, sizeof (iInt)) < 0)
        log_message ("feX", "setsockopt error");

    if (bind (iLocalFD, (struct sockaddr *) &sLocalSocket, sizeof (sLocalSocket)) < 0)
        log_message ("feX", "listen error");

    if (listen (iLocalFD, nBacklogSize) < 0)
        log_message ("feX", "listen error");

    memset (&sSigAction, 0, sizeof (sSigAction));
    sSigAction.sa_handler = child_signal;
    if (sigaction (SIGCHLD, &sSigAction, 0))
        log_message ("feX", "signal error");

    if (setvbuf (stdout, NULL, _IONBF, 0) != 0)
      log_message ("fe", "warning: unable to disable buffer stdout");

    for (;;)
    {
        iSocketSize = sizeof (sClientSocket);
        if ((iConnectionFD = accept (iLocalFD, (struct sockaddr *) &sClientSocket, &iSocketSize)) < 0)
        {
            if (errno == EINTR || errno == ECONNABORTED)
                continue;
            else
                log_message ("feX", "accept connection error");
        }

        if ((iChildPID = fork()) == -1)
            log_message ("feX", "fork error");

        if (iChildPID == 0)
            branch_tcp (iLocalFD, iConnectionFD, sOpts.iTCPerrors, &sClientSocket, &sConfig);

        if ((close (iConnectionFD)) == -1)
            log_message ("feX", "close connection error"); /* this aborts too */

    }

    return EXIT_SUCCESS; /* unreachable anyway */
}
